import { Post, Tag, LogEntry, LogType } from "../type/generic";
import { Scrapper } from "../class/Scrapper";
import { getPageContents } from "../helper/requestManager";
import * as cheerio from 'cheerio';

export class Rule34xxx extends Scrapper {

    constructor() {
        // Set the domain base of the current Scrapper as "rule34.xxx"
        super("https://rule34.xxx");
    }

    /**
     * Get the details of a specific post
     * @param url The URL to the post, this must be the actual page which contains the image, tags, etc...
     */
    public async getPostDetails(url: string): Promise<Post | null> {

        // Check if the provided link is valid
        if (!this.checkURLBase(url)) {
            throw new Error(`Invalid url provided`);
        }

        // Init the variable here in case of an error
        let pageContents = null;
        
        // Send out the request to grab the contents of the post
        try {
            // Send out the initial Axios request to fetch the data from the page
            pageContents = await getPageContents(url);

            if (pageContents.status < 200 || pageContents.status > 299) {
                throw new Error(`Invalid response code[${pageContents.status}]`);
            }

            pageContents = pageContents.data;

        } catch (err) {
            // "Handle" the error so that it's in the above .catch
            if (this.verbose) {
                console.error(`[Error]::getPostDetails::`);
            }
            throw err;
        }

        // Process the page's posts with cheerio
        const $ = cheerio.load((pageContents as string));

        const postTags: Array<Tag> = [];
        { // Get the post's tags
            const tagsSection = $(`#tag-sidebar`);
            if (tagsSection.length <= 0) {
                throw new Error(`Failed to find post tags, invalid post`);
            }

            // Run a query for all tags
            tagsSection.find(`.tag`).each(function () {
                // Go through the classes of the tag and see if we find the type
                let tagType = "general";
                const classList = $(this).attr(`class`).split(" ");
                if (classList.length > 0) {
                    for (let tt of classList) {
                        if (tt.includes(`tag-type-`)) {
                            tagType = tt.split(`-type-`)[1];
                            break;
                        }
                    }
                }

                const tagAnchor = $(this).find(`a`);
                let tagName = `unknown`;
                let tagSlug = `unknown`;
                { // Get the name of the tag and slug
                    if (tagAnchor.length > 0) {
                        tagName = tagAnchor.text();
                        tagSlug = tagAnchor.attr(`href`).split("tags=")[1];
                    }
                }

                // Add the tag to the postTags listing
                postTags.push({
                    slug: tagSlug ?? `unknown`,
                    type: tagType ?? `general`
                });

            })

        }

        let postContent = `ERROR`;
        { // get the link to the post's original image/video
            const imageLink = $(`meta[property="og:image"]`);

            if (imageLink.length > 0) {
                postContent = imageLink.attr(`content`);
            }

            // Make sure the postContent isn't just a link back like they like to do and/or we didn't find anything
            if (postContent == `ERROR` || postContent == url) {
                // Get the current page's contnet
                postContent = $(`#fit-to-screen img`).attr(`src`);
            }

            if (postContent.indexOf(`?`) >= 5) {
                postContent = postContent.split(`?`)[0];
            }

            if (postContent.indexOf(`//`) >= 0) {
                postContent = postContent;
            }

        }

        // Get the source of the post
        let postSource = null;
        {
            const sourceA = $(`#stats a[rel="nofollow"]`);
            if (sourceA.length > 0) {
                postSource = sourceA.attr(`href`);
            }
        }

        let postDate = "2021-10-17 13:18:27";
        {
            const postDateRef = $(`#stats li:nth-child(2)`);
            if (postDateRef.length > 0) {
                postDate = postDateRef.text().split("\n")[1].replace(/Posted: /g, '');
            }
        }

        return {
            url: url,
            contentURL: postContent ?? "ERROR",
            source: postSource,
            tags: postTags ?? [],
            ts: postDate,
        };
    }

    /**
     * Get a list of posts from the mentioned page
     * @param url 
     * @returns 
     */
    public async getPostsFromPage(url: string): Promise<Array<string>> {

        // Check if the provided link is valid
        if (!this.checkURLBase(url)) {
            throw new Error(`Invalid url provided`);
        }

        // Initialize the page contents here
        let pageContents: string = null;

        // Send out the request to grab the contents of the post
        try {
            if (this.verbose) {
                console.error(`Sniffing page...`);
            }
            // Send out the initial Axios request to fetch the data from the page
            await getPageContents(url)
                .then(request => {
                    if (request.status < 200 || request.status > 299) {
                        this.logs.push({
                            msg: `Invalid response code[${request.status}]`,
                            type: LogType.ERROR,
                            err: null,
                            data: null,
                            ts: new Date()
                        });
                        throw new Error(`Invalid response code[${request.status}]`);
                    }

                    pageContents = (request.data as string);
                })
        } catch (err) {
            // "Handle" the error so that it's in the above .catch
            this.logs.push({
                msg: `[Error]::getPostsFromPage::`,
                type: LogType.ERROR,
                err: (err as Error),
                data: null,
                ts: new Date()
            });
            throw err;
        }

        // Process the page's posts with cheerio
        const $ = cheerio.load((pageContents as string));

        // Define the post List
        const postList: Array<string> = [];

        // Workaround I guess
        let self = this;

        // Go through all of the posts
        $(`.thumb`).each(function () {
            const href = $(this).find(`a`).attr(`href`);
            if (`${href}`.length >= `index.php?page=post&s=view&id=`.length)
                postList.push(`${self.domain}/${href}`);
        });

        if (this.verbose) {
            console.error(`Found ${postList.length} posts`);
        }

        return postList;
    }

    public async crawlPages(url: string, pageCount: number = 10, batchSize: number = 0): Promise<Array<string>> {

        // Check if the provided link is valid
        if (!this.checkURLBase(url)) {
            throw new Error(`Invalid url provided`);
        }

        // A list of all of the found pages
        let foundPages = new Array<string>();
        // The next url we are hitting
        let nextPage: string = url;

        // Go through as many pages as requested
        for (let i = 0; i < pageCount; i++) {

            if (this.verbose) {
                console.log(`[${i + 1}/${pageCount}]Crawling ${nextPage}`);
            }

            // Initialize the page contents here
            let pageContents: string = null;

            // Try and find the button to the next page
            try {
                // Send out the initial Axios request to fetch the data from the page
                await getPageContents(nextPage)
                    .then(request => {
                        if (request.status < 200 || request.status > 299) {
                            this.logs.push({
                                msg: `Invalid response code[${request.status}]`,
                                type: LogType.ERROR,
                                err: null,
                                data: null,
                                ts: new Date()
                            });
                            throw new Error(`Invalid response code[${request.status}]`);
                        }

                        pageContents = (request.data as string);
                    })


            } catch (err) {
                // "Handle" the error so that it's in the above .catch
                this.logs.push({
                    msg: `[Error]::getPostsFromPage::`,
                    type: LogType.ERROR,
                    err: (err as Error),
                    data: null,
                    ts: new Date()
                });
                throw err;
            }

            // Process the page's posts with cheerio
            const $ = cheerio.load((pageContents as string));

            // Add the current page we are on to the list
            foundPages.push(nextPage);

            const nextPageButton = $(`a[alt="next"]`);
            if (nextPageButton.length > 0) {
                nextPage = `${this.domain}/` + nextPageButton.attr(`href`);
            } else {
                // Since we didn't find the proper button, skip this page.
                break;
            }
        }

        // Return the found pages
        return foundPages;
    }

}